from dataclasses import dataclass

from sharded.schema import BucketStatus, StorageClass, StorageDriver


@dataclass
class Entity:
    id: int


@dataclass
class Storage(Entity):
    storage_class: StorageClass
    driver: StorageDriver
    dsn: str

    def __hash__(self) -> int:
        return hash(self.id)


@dataclass
class Bucket(Entity):
    key: str
    repository: str
    storage_id: int
    status: BucketStatus

    def __hash__(self) -> str:
        return hash((self.repository, self.key))



from dataclasses import dataclass
from functools import cache
from typing import Optional

from sharded.drivers import Driver
from sharded.entity import Bucket, Entity, Storage
from sharded.schema import BucketStatus


@dataclass
class Index:
    entity: type[Entity]
    fields: list[str]
    unique: bool = False


class UniqueIndex(Index):
    unique = True


class Repository:
    entities: list[type[Entity]]
    indexes: Optional[list[Index]] = None

    def __init__(self) -> None:
        if not self.indexes:
            self.indexes = []

        for entity in self.entities:
            self.indexes.insert(0, UniqueIndex(entity, ['id']))

    async def cast_storage(self, storages: list[Storage]) -> Storage:
        return storages[0]

    async def get_key(self, context: dict) -> str:
        return ''

    async def init_data(self, bucket: Bucket, driver: Driver) -> None:
        bucket.status = BucketStatus.READY

    async def init_schema(self, driver: Driver) -> None:
        for entity in self.entities:
            await driver.init_schema(entity)

    def make(self, entity: type[Entity], row: dict) -> Entity:
        return entity(**{k: v for (k, v) in row.items() if k != 'bucket_id'})


@cache
def get_entity_repository_class(entity: type[Entity]) -> type[Repository]:
    repositories = [
        repository for repository in Repository.__subclasses__()
        if entity in repository.entities
    ]
    if not len(repositories):
        raise LookupError(f'No entity repository found: {entity}')

    if len(repositories) > 1:
        raise LookupError(f'Duplicate entity repository: {entity}')

    return repositories[0]


class BucketRepository(Repository):
    entities = [Bucket]

    async def bootstrap(self, driver: Driver) -> tuple[Bucket, Bucket]:
        bucket_row = await driver.find_or_create(
            name='Bucket',
            find={'id': 1},
            create={
                'bucket_id': 1,
                'id': 1,
                'key': '',
                'repository': BucketRepository,
                'status': BucketStatus.READY,
                'storage_id': 1,
            }
        )

        storage_row = await driver.find_or_create(
            name='Bucket',
            find={'id': 2},
            create={
                'bucket_id': 1,
                'id': 2,
                'key': '',
                'repository': StorageRepository,
                'status': BucketStatus.READY,
                'storage_id': 1,
            }
        )

        self.bucket_bucket = self.make(Bucket, bucket_row)
        self.storage_bucket = self.make(Bucket, storage_row)


class StorageRepository(Repository):
    entities = [Storage]

    async def bootstrap(self, driver: Driver, storage: Storage):
        await driver.find_or_create(
            name='Storage',
            find={'id': storage.id},
            create=dict(bucket_id=2, **storage.__dict__),
        )



from functools import cache
from typing import Protocol
from sharded.entity import Entity

from sharded.schema import StorageDriver


class Driver(Protocol):
    def __init__(self, dsn: str) -> None:
        ...

    async def find(self, name: str, params: dict) -> list[dict]:
        raise NotImplementedError()

    async def find_or_create(
        self, name: str, find: dict, create: dict
    ) -> dict:
        raise NotImplementedError()

    async def init_schema(self, entity: Entity) -> None:
        raise NotImplementedError()


driver_instances: dict[str, dict[str, Driver]] = {}


async def get_driver(driver: StorageDriver, dsn: str) -> Driver:
    if driver not in driver_instances:
        driver_instances[driver] = {}
    if dsn not in driver_instances[driver]:
        driver_instances[driver][dsn] = get_implementation(driver)(dsn)
    return driver_instances[driver][dsn]


@cache
def get_implementation(driver):
    implementations: dict[StorageDriver, Driver] = {
        StorageDriver.MEMORY: MemoryDriver
    }
    if driver in implementations:
        return implementations[driver]
    raise NotImplementedError(f'Driver {driver} not implemented')


class MemoryDriver(Driver):
    def __init__(self, dsn: str) -> None:
        super().__init__(dsn)
        self.data: dict[str, list[dict]] = {}

    async def find_or_create(
        self, name: str, find: dict, create: dict
    ) -> dict:
        rows = await self.find(name, [find])
        if len(rows):
            return rows[0]

        create['id'] = len(self.data[name]) + 1
        self.data[name].append(create)
        return create

    async def find(self, name: str, params: list) -> list[dict]:
        return [
            row for row in self.data[name]
            if await self.is_valid(row, params)
        ]

    async def is_valid(self, row, params) -> bool:
        for param in params:
            if False not in [row[key] == value for (key, value) in param.items()]:
                return True

        return False

    async def init_schema(self, entity: Entity) -> None:
        if entity.__name__ not in self.data:
            self.data[entity.__name__] = []



from enum import StrEnum


class BucketStatus(StrEnum):
    NEW = 'n'
    SCHEMA = 's'
    READY = 'r'


class StorageClass(StrEnum):
    COLD = 'cold'
    HOT = 'hot'
    ICE = 'ice'
    MEMORY = 'memory'


class StorageDriver(StrEnum):
    MEMORY = 'memory'
    TARANTOOL = 'tarantool'



from dataclasses import dataclass
from functools import cache
from typing import Optional, TypeVar

from sharded.drivers import Driver, get_driver
from sharded.entity import Bucket, Entity, Storage
from sharded.repository import (BucketRepository, Repository,
                                     StorageRepository,
                                     get_entity_repository_class)
from sharded.schema import BucketStatus, StorageClass, StorageDriver


@dataclass
class QueryContext:
    bucket: Bucket
    driver: Driver
    entity: Entity
    repository: Repository


class Gateway:
    def __init__(self) -> None:
        self.ready: bool = False
        self.repositories: dict[type[Repository], Repository] = {
            repository: repository()
            for repository in Repository.__subclasses__()
        }
        self.storages: list[Storage] = [
            Storage(
                id=1,
                storage_class=StorageClass.MEMORY,
                driver=StorageDriver.MEMORY,
                dsn=''
            )
        ]

    async def bootstrap(self, source: Optional[Storage] = None):
        if self.ready:
            return

        self.ready = True
        if not source:
            source = self.storages[0]

        driver = await get_driver(source.driver, source.dsn)
        await driver.init_schema(Bucket)
        await driver.init_schema(Storage)
        await self.repositories[BucketRepository].bootstrap(driver)
        await self.repositories[StorageRepository].bootstrap(driver, source)

    async def find_or_create(
        self,
        entity: type[Entity],
        create: Optional[dict | list] = None,
        find: Optional[dict | list] = None,
        key: Optional[any] = None,
    ) -> Entity:
        if find is None:
            find = create
        context = await self.context(entity, key)
        return context.repository.make(
            entity=entity,
            row=await context.driver.find_or_create(
                name=entity.__name__,
                find=dict(bucket_id=context.bucket.id, **find),
                create=dict(bucket_id=context.bucket.id, **create),
            )
        )

    async def find(
        self,
        entity: type[Entity],
        params: Optional[dict | list] = None,
        key: Optional[any] = None,
    ) -> list[Entity]:
        # return repository.get_instances(entity, data)
        # return await repository.find(entity, bucket, params)
        context = await self.context(entity, key)
        if not params:
            params = {}
        if isinstance(params, dict):
            params = [params]

        params = [
            dict(bucket_id=context.bucket.id, **param) for param in params
        ]
        rows = await context.driver.find(entity.__name__, params)

        return [context.repository.make(entity, row) for row in rows]

    async def get(
        self,
        entity: type[Entity],
        id: int,
        key: Optional[any] = None,
    ) -> Optional[Entity]:
        instances = await self.find(entity, {'id': id}, key)
        if len(instances):
            return instances[0]

    async def context(self, entity: type[Entity], key: any) -> QueryContext:
        await self.bootstrap()

        repository = self.repositories[get_entity_repository_class(entity)]
        bucket = await self.get_bucket(repository, key)

        if not bucket.storage_id:
            storage = await repository.cast_storage(self.storages)
            bucket.storage_id = storage.id
        else:
            storage = self.get_storage(bucket.storage_id)

        driver = await get_driver(storage.driver, storage.dsn)

        if bucket.status == BucketStatus.NEW:
            await repository.init_schema(driver)
            bucket.status = BucketStatus.SCHEMA
            # await self.persist(bucket)

        if bucket.status == BucketStatus.SCHEMA:
            await repository.init_data(bucket, driver)
            # await self.persist(bucket)

        if bucket.status != BucketStatus.READY:
            raise LookupError(f'Invalid status: {bucket.status}')

        return QueryContext(bucket, driver, entity, repository)

    async def get_bucket(self, repository: Repository, key: any) -> Bucket:
        if isinstance(repository, BucketRepository):
            return self.repositories[BucketRepository].bucket_bucket
        if isinstance(repository, StorageRepository):
            return self.repositories[BucketRepository].storage_bucket
        return await self.find_or_create(
            entity=Bucket,
            find={
                'key': await repository.get_key(key),
                'repository': repository,
            },
            create={
                'key': await repository.get_key(key),
                'repository': repository,
                'status': BucketStatus.NEW,
                'storage_id': 0,
            },
        )

    @cache
    def get_storage(self, storage_id: int) -> Storage:
        storage = None
        for candidate in self.storages:
            if candidate.id == storage_id:
                storage = candidate

        if not Storage:
            raise LookupError(
                f'storage {storage_id} not found'
            )

        return storage



from dataclasses import dataclass

from pytest import mark

from sharded.entity import Entity
from sharded.gateway import Gateway
from sharded.repository import Index, Repository


@dataclass
class Action(Entity):
    type: str


@dataclass
class ActionTrigger(Entity):
    ...


class ActionRepository(Repository):
    entities = [
        Action,
        ActionTrigger
    ]
    indexes = [
        Index(ActionTrigger, ['id'])
    ]


@mark.asyncio
async def test_hello():
    gateway = Gateway()
    assert len(await gateway.find(Action)) == 0
    action1 = await gateway.find_or_create(Action, {'type': 'tester'})
    action2 = await gateway.find_or_create(Action, {'type': 'tester2'})
    assert action1.id == 1
    assert action1.type == 'tester'
    assert action2.id == 2
    assert action2.type == 'tester2'
    assert len(await gateway.find(Action)) == 2
    assert (await gateway.get(Action, 2)).type == 'tester2'
    assert (await gateway.get(Action, 3)) is None
